from time import sleep
from typing import Callable, Any

from click_extra import progressbar

from geocompy.geo import GeoCom
from geocompy.communication import open_serial

from ..utils import (
    echo_red,
    echo_green
)


MORSE_TABLE = {
    "a": ".-",
    "b": "-...",
    "c": "-.-.",
    "d": "-..",
    "e": ".",
    "f": "..-.",
    "g": "--.",
    "h": "....",
    "i": "..",
    "j": ".---",
    "k": "-.-",
    "l": ".-..",
    "m": "--",
    "n": "-.",
    "o": "---",
    "p": ".--.",
    "q": "--.-",
    "r": ".-.",
    "s": "...",
    "t": "-",
    "u": "..-",
    "v": "...-",
    "w": ".--",
    "x": "-..-",
    "y": "-.--",
    "z": "--..",
    "1": ".----",
    "2": "..---",
    "3": "...--",
    "4": "....-",
    "5": ".....",
    "6": "-....",
    "7": "--...",
    "8": "---..",
    "9": "----.",
    "0": "-----",
    "&": ".-...",
    "'": ".----.",
    "@": ".--.-.",
    "(": "-.--.",
    ")": "-.--.-",
    ":": "---...",
    ",": "--..--",
    "=": "-...-",
    "!": "-.-.--",
    ".": ".-.-.-",
    "-": "-....-",
    "%": "------..-.-----",
    "+": ".-.-.",
    "\"": ".-..-.",
    "?": "..--..",
    "/": "-..-.",
    "\n": ".-.-"
}


def encode_message(
    message: str
) -> str:
    words: list[str] = []
    for word in message.lower().split(" "):
        w: list[str] = []
        for letter in word:
            w.append("|".join(MORSE_TABLE.get(letter, "")))

        words.append("_".join(w))

    return " ".join(words)


def relay_message(
    beepstart: Callable[[], Any],
    beepstop: Callable[[], Any],
    message: str,
    unittime: float
) -> None:
    encoded = encode_message(message)
    with progressbar(
        encoded,
        label="Relaying message",
        show_eta=False
    ) as stream:
        for char in stream:
            match char:
                case ".":
                    beepstart()
                    sleep(unittime)
                    beepstop()
                case "-":
                    beepstart()
                    sleep(3 * unittime)
                    beepstop()
                case "|":
                    sleep(unittime)
                case "_":
                    sleep(3 * unittime)
                case " ":
                    sleep(7 * unittime)
                case _:
                    raise ValueError(
                        f"Invalid morse stream character: '{char}'"
                    )


def main(
    port: str,
    message: str,
    intensity: int = 100,
    ignore_non_ascii: bool = False,
    baud: int = 9600,
    timeout: int = 15,
    retry: int = 1,
    sync_after_timeout: bool = False,
    unittime: int = 50,
    compatibility: str = "none",
) -> None:
    if not ignore_non_ascii:
        try:
            message.casefold().encode("ascii")
        except UnicodeEncodeError:
            echo_red("The message contains non-ASCII characters.")
            exit(1)

    with open_serial(
        port,
        speed=baud,
        timeout=timeout,
        retry=retry,
        sync_after_timeout=sync_after_timeout
    ) as com:
        tps = GeoCom(com)
        beepstart = tps.bmm.beep_start
        beepstop = tps.bmm.beep_stop
        match compatibility.lower():
            case "tps1000":
                beepstart = tps.bmm.beep_on
                beepstop = tps.bmm.beep_off
            case "none":
                pass

        relay_message(
            lambda: beepstart(intensity),
            beepstop,
            message,
            unittime * 1e-3
        )
        echo_green("Message complete.")



from logging import DEBUG, ERROR, INFO, WARNING, Logger
import os
from typing import Any, Callable, cast, TypeVar

from click_extra import (
    Color,
    echo,
    style,
    option,
    option_group,
    argument,
    Choice,
    IntRange
)
from cloup.constraints import mutually_exclusive


F = TypeVar('F', bound=Callable[..., Any])


EXIT_CODE_DESCRIPTIONS: dict[int, str] = {
    1: "Unknown",
    2: "Keyboard interrupt",
    3: "Missing dependencies",
    4: "Malformed data",
    1100: "Error in target point CSV",
    1101: "Duplicate targets between CSV and existing JSON",
    1102: "Error while opening point CSV",
    1103: "Target CSV file does not exist",
    1200: "Unknown measurement order"
}


def com_port_argument() -> Callable[[F], F]:
    return argument(
        "port",
        help=(
            "serial port that the instrument is connected to (must be a valid "
            "identifier like COM1 or /dev/usbtty0)"
        ),
        type=str
    )


def com_timeout_option(
    default: int = 15
) -> Callable[[F], F]:
    return option(
        "-t",
        "--timeout",
        help="serial timeout",
        type=IntRange(min=0),
        default=default
    )


def com_baud_option(
    default: int = 9600
) -> Callable[[F], F]:
    return option(
        "-b",
        "--baud",
        help="serial speed",
        type=Choice(
            [
                "1200",
                "2400",
                "4800",
                "9600",
                "19200",
                "38400",
                "56000",
                "57600",
                "115200",
                "230400",
                "921600"
            ]
        ),
        callback=lambda ctx, param, value: int(value),
        default=str(default)
    )


def com_option_group() -> Callable[[F], F]:
    return option_group(
        "Connection options",
        "Options related to the serial connection",
        com_baud_option(),
        com_timeout_option(),
        option(
            "-r",
            "--retry",
            help="number of connection retry attempts",
            type=IntRange(min=0, max=10),
            default=1
        ),
        option(
            "--sync-after-timeout",
            help="attempt to synchronize message que after a timeout",
            is_flag=True
        )
    )


def logging_option_group() -> Callable[[F], F]:
    return option_group(
        "Logging options",
        "Options related to the logging functionalities.",
        option(
            "--debug",
            is_flag=True
        ),
        option(
            "--info",
            is_flag=True
        ),
        option(
            "--warning",
            is_flag=True
        ),
        option(
            "--error",
            is_flag=True
        ),
        constraint=mutually_exclusive
    )


def echo_color(
    message: Any,
    color: str,
    newline: bool = True,
    error: bool = False
) -> None:
    echo(
        style(
            message,
            color
        ),
        nl=newline,
        err=error
    )


def echo_yellow(
    message: Any,
    newline: bool = True,
    error: bool = False
) -> None:
    echo_color(message, Color.yellow, newline, error)


def echo_green(
    message: Any,
    newline: bool = True,
    error: bool = False
) -> None:
    echo_color(message, Color.green, newline, error)


def echo_red(
    message: Any,
    newline: bool = True,
    error: bool = False
) -> None:
    echo_color(message, Color.red, newline, error)


def make_directory(filepath: str) -> None:
    dirname = os.path.dirname(filepath)
    if dirname == "":
        return

    os.makedirs(dirname, exist_ok=True)


def make_logger(
    name: str,
    debug: bool = False,
    info: bool = False,
    warning: bool = False,
    error: bool = False
) -> Logger:
    from geocompy.communication import get_logger

    if debug:
        loglevel = DEBUG
    elif info:
        loglevel = INFO
    elif warning:
        loglevel = WARNING
    elif error:
        loglevel = ERROR
    else:
        return get_logger(name)

    return get_logger(name, "stdout", loglevel)


def run_cli_app(
    name: str,
    runner: Callable[..., Any],
    *args: Any
) -> None:
    logger = make_logger("APP", info=True)
    try:
        logger.info(f"Starting '{name}' application")
        runner(args)
    except KeyboardInterrupt:
        logger.info("Keyboard interrupt...")
        exit(2)
    except SystemExit as ex:
        if ex.code == 0:
            logger.info(f"Application '{name}' exited without error")
            raise ex

        logger.error(
            f"Application exited with {ex.code} "
            f"({EXIT_CODE_DESCRIPTIONS.get(cast(int, ex.code), 'Unknown')})"
        )
        raise ex
    except Exception:
        logger.exception(
            f"Application '{name}' exited due to an unhandled exception"
        )
        exit(1)

    logger.info(f"Application '{name}' finished without error")
    exit(0)



from pathlib import Path
from typing import Callable, Any
from enum import Enum

from geocompy.data import Angle
from geocompy.communication import open_serial
from geocompy.geo import GeoCom
from geocompy.geo.gctypes import GeoComResponse, GeoComSubsystem, GeoComCode
from geocompy.gsi.dna import GsiOnlineDNA
from geocompy.gsi.gsitypes import GsiOnlineResponse

from ..utils import echo_red, echo_green
from .io import write_settings, SettingsDict, SubsystemSettingsDict


def download_settings_geocom(
    tps: GeoCom,
    defaults: bool = False
) -> SettingsDict:
    options: list[SubsystemSettingsDict] = [
        {
            "subsystem": "aut",
            "options": {
                "atr": True,
                "lock": False,
                "tolerance": [1, 1],
                "timeout": [15, 15],
                "fine_adjust_mode": "NORMAL",
                "search_area": [0, 0, 0.1, 0.1, True],
                "spiral": [0.1, 0.1],
                "lock_onthefly": True
            }
        },
        {
            "subsystem": "bap",
            "options": {
                "target_type": "REFLECTOR",
                "prism_type": "MINI",
                "measurement_program": "SINGLE_REF_STANDARD",
                "atr_setting": "NORMAL",
                "reduced_atr_fov": False,
                "precise_atr": True
            }
        },
        {
            "subsystem": "csv",
            "options": {
                "laserlot": True,
                "laserlot_intensity": 100,
                "charging": False,
                "preferred_powersource": "INTERNAL"
            }
        },
        {
            "subsystem": "dna",
            "options": {
                "staffmode": False,
                "curvature_correction": False,
                "staff_type": "GPCL2"
            }
        },
        {
            "subsystem": "edm",
            "options": {
                "laserpointer": False,
                "edm": True,
                "boomerang_filter": True,
                "tracklight_brightness": "MID",
                "tracklight": False,
                "guidelight_intensity": "OFF",
                "boomerang_filter_new": True
            }
        },
        {
            "subsystem": "img",
            "options": {
                "telescopic_configuration": [0, 50, 6, "INTERNAL"],
                "telescopic_exposure_time": 20
            }
        },
        {
            "subsystem": "kdm",
            "options": {
                "display_power": False
            }
        },
        {
            "subsystem": "sup",
            "options": {
                "poweroff_configuration": [False, "SLEEP", 600000],
                "low_temperature_control": False,
                "autorestart": True
            }
        },
        {
            "subsystem": "tmc",
            "options": {
                "compensator": True,
                "edm_mode_v1": "SINGLE_STANDARD",
                "edm_mode_v2": "SINGLE_STANDARD",
                "angle_correction": [True, True, True, True]
            }
        }
    ]
    output_options: list[SubsystemSettingsDict] = []
    for group in options:
        settings: SubsystemSettingsDict = {
            "subsystem": group["subsystem"],
            "options": {}
        }

        subsystem: GeoComSubsystem | None = getattr(
            tps,
            group["subsystem"],
            None
        )
        if subsystem is None:
            continue

        for option, default in group["options"].items():
            if isinstance(default, bool):
                name = f"get_{option}_status"
            else:
                name = f"get_{option}"

            method: Callable[
                [],
                GeoComResponse[Any]
            ] | None = getattr(subsystem, name, None)
            if method is None:
                settings["options"][option] = default if defaults else None
                continue

            response = method()
            value = response.params
            if response.error != GeoComCode.OK or value is None:
                settings["options"][option] = default if defaults else None
                continue

            if isinstance(value, tuple):
                raw = value
                value = []
                for v in raw:
                    if isinstance(v, Enum):
                        value.append(v.name)
                    elif isinstance(v, Angle):
                        value.append(float(v))
                    else:
                        value.append(v)
            elif isinstance(value, Enum):
                value = value.name

            settings["options"][option] = value

        output_options.append(settings)

    return {
        "protocol": "geocom",
        "settings": output_options
    }


def download_settings_gsidna(
    dna: GsiOnlineDNA,
    defaults: bool = False
) -> SettingsDict:
    settings: SubsystemSettingsDict = {
        "subsystem": "settings",
        "options": {}
    }

    options = {
        "beep": "MEDIUM",
        "contrast": 50,
        "distance_unit": "METER",
        "temperature_unit": "CELSIUS",
        "decimals": 5,
        "baud": "B9600",
        "parity": "NONE",
        "terminator": "CRLF",
        "protocol": True,
        "recorder": "INTERNAL",
        "delay": 0,
        "autooff": "SLEEP",
        "display_heater": False,
        "curvature_correction": False,
        "staff_mode": False,
        "format": "GSI8",
        "code_recording": "BEFORE"
    }

    for option, default in options.items():
        name = f"get_{option}"
        method: Callable[
            [],
            GsiOnlineResponse[Any]
        ] | None = getattr(dna.settings, name, None)
        if method is None:
            settings["options"][option] = default if defaults else None
            continue

        response = method()
        value = response.value
        if value is None:
            settings["options"][option] = default if defaults else None
            continue

        if isinstance(value, Enum):
            value = value.name

        settings["options"][option] = value

    return {
        "protocol": "gsidna",
        "settings": [settings]
    }


def clean_settings(
    settings: SettingsDict
) -> SettingsDict:
    for subsystem in settings["settings"]:
        subsystem["options"] = {
            k: v for k, v in subsystem["options"].items() if v is not None
        }

    settings["settings"] = [
        s
        for s in settings["settings"]
        if len(s["options"]) > 0
    ]

    return settings


def main(
    port: str,
    protocol: str,
    file: Path,
    baud: int = 9600,
    timeout: int = 15,
    retry: int = 1,
    sync_after_timeout: bool = False,
    format: str = "auto",
    defaults: bool = False
) -> None:
    with open_serial(
        port,
        retry=retry,
        sync_after_timeout=sync_after_timeout,
        speed=baud,
        timeout=timeout
    ) as com:
        match protocol:
            case "geocom":
                tps = GeoCom(com)
                data = download_settings_geocom(tps, defaults)
            case "gsidna":
                dna = GsiOnlineDNA(com)
                data = download_settings_gsidna(dna, defaults)
            case _:
                echo_red(f"Unknown protocol: '{protocol}'")
                exit(1)

    data = clean_settings(data)

    write_settings(data, file, format)
    echo_green(f"Settings saved at {file}")



from typing import Any

from click_extra import (
    extra_command,
    option,
    argument,
    Choice,
    IntRange
)

from ..utils import (
    com_option_group,
    com_port_argument
)


@extra_command(
    "morse",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@argument(
    "message",
    help="message to relay as a string of ASCII characters",
    type=str
)
@option(
    "-i",
    "--intensity",
    help="beeping intensity",
    type=IntRange(0, 100),
    default=100
)
@option(
    "-u",
    "--unittime",
    help="beep unit time in milliseconds [ms]",
    type=IntRange(min=50),
    default=50
)
@option(
    "-c",
    "--compatibility",
    help="instrument compatibility",
    type=Choice(["none", "TPS1000"], case_sensitive=False),
    default="none"
)
@option(
    "--ignore-non-ascii",
    help="suppress encoding errors and skip non-ASCII characters",
    is_flag=True
)
@com_option_group()
def cli(**kwargs: Any) -> None:
    """Play a Morse encoded ASCII message through the beep signals
        of a GeoCom capable total station.
        """

    from .app import main

    main(**kwargs)



from click_extra import echo
from geocompy.communication import open_serial
from geocompy.geo import GeoCom
from geocompy.geo.gctypes import GeoComCode

from ..utils import echo_red, echo_yellow


def run_listing(tps: GeoCom) -> None:
    resp_setup = tps.csv.setup_listing()
    if resp_setup.error != GeoComCode.OK:
        echo_red(f"Could not set up job listing ({resp_setup.error.name})")
        return

    resp_list = tps.csv.list()
    if resp_list.error != GeoComCode.OK or resp_list.params is None:
        echo_red(f"Could not start listing ({resp_list.error.name})")
        return

    job, file, _, _, _ = resp_list.params
    if job == "" or file == "":
        echo_yellow("No jobs were found")
        return

    count = 1
    echo(f"{'job name':<50.50s}{'file name':<50.50s}")
    echo(f"{'--------':<50.50s}{'---------':<50.50s}")
    fmt = "{job:<50.50s}{file:<50.50s}"
    echo(
        fmt.format_map(
            {
                "job": job,
                "file": file
            }
        )
    )
    while True:
        resp_list = tps.csv.list()
        if resp_list.error != GeoComCode.OK or resp_list.params is None:
            break

        job, file, _, _, _ = resp_list.params
        echo(
            fmt.format_map(
                {
                    "job": job,
                    "file": file
                }
            )
        )
        count += 1

    echo("-" * 90)
    echo(f"total: {count} files")


def main_list(
    port: str,
    baud: int = 9600,
    timeout: int = 15,
    retry: int = 1,
    sync_after_timeout: bool = False
) -> None:
    with open_serial(
        port=port,
        speed=baud,
        timeout=timeout,
        retry=retry,
        sync_after_timeout=sync_after_timeout
    ) as com:
        tps = GeoCom(com)
        run_listing(tps)



from typing import Any

from click_extra import (
    extra_command,
    argument,
    option,
    IntRange,
    File
)

from ..utils import (
    com_port_argument,
    com_baud_option,
    com_timeout_option
)


@extra_command(
    "data",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@com_baud_option()
@com_timeout_option(2)
@option(
    "-o",
    "--output",
    help="file to save received data",
    type=File("wb", encoding="utf8", lazy=True)
)
@option(
    "--eof",
    help="end-of-file marker (i.e. the last line to receive)",
    type=str,
    default=""
)
@option(
    "--autoclose/--no-autoclose",
    help="close transfer automatically upon timeout or when EOF is received",
    default=True
)
@option(
    "--inclide-eof/--no-include-eof",
    help=(
        "wether the EOF marker is part of the output format "
        "(or just sent by the instrument regardless of the format in question)"
    ),
    default=False
)
def cli_download(**kwargs: Any) -> None:
    """Receive data sent from the instrument."""
    from .app import main_download

    main_download(**kwargs)


@extra_command(
    "data",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@argument(
    "file",
    help="data file to upload",
    type=File("rt", encoding="ascii")
)
@com_baud_option(1200)
@com_timeout_option()
@option(
    "-s",
    "--skip",
    help="number of header rows to skip",
    type=IntRange(min=0),
    default=0
)
def cli_upload(**kwargs: Any) -> None:
    """Upload ASCII data to the instrument."""
    from .app import main_upload

    main_upload(**kwargs)



from typing import Any

from click_extra import (
    extra_command,
    option,
    argument,
    IntRange,
    Choice,
    File
)

from ..utils import (
    com_option_group,
    com_port_argument
)


@extra_command(
    "files",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@argument(
    "directory",
    help="directory to list files in (path should end with '/')",
    type=str,
    default="/"
)
@com_option_group()
@option(
    "-d",
    "--device",
    help="memory device",
    type=Choice(
        (
            "internal",
            "cf",
            "sd",
            "usb",
            "ram"
        ),
        case_sensitive=False
    ),
    default="internal"
)
@option(
    "-f",
    "--filetype",
    help="file type",
    type=Choice(
        (
            "image",
            "database",
            "overview-jpg",
            "overview-bmp",
            "telescope-jpg",
            "telescope-bmp",
            "scan",
            "unknown",
            "last"
        ),
        case_sensitive=False
    ),
    default="unknown"
)
def cli_list(**kwargs: Any) -> None:
    """List files on an instrument."""
    from .app import main_list

    main_list(**kwargs)


@extra_command(
    "file",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@argument(
    "filename",
    help=(
        "file to download (including path with '/' separators if filetype "
        "option is not specified)"
    ),
    type=str
)
@argument(
    "output",
    help="file to save downloaded data to",
    type=File("wb", lazy=False)
)
@com_option_group()
@option(
    "-d",
    "--device",
    help="memory device",
    type=Choice(
        (
            "internal",
            "cf",
            "sd",
            "usb",
            "ram"
        ),
        case_sensitive=False
    ),
    default="internal"
)
@option(
    "-f",
    "--filetype",
    help="file type",
    type=Choice(
        (
            "image",
            "database",
            "overview-jpg",
            "overview-bmp",
            "telescope-jpg",
            "telescope-bmp",
            "scan",
            "unknown",
            "last"
        ),
        case_sensitive=False
    ),
    default="unknown"
)
@option(
    "-c",
    "--chunk",
    help="chunk size (max 450 for normal and 1800 for VivaTPS large download)",
    type=IntRange(1, 1800),
    default=450
)
@option(
    "--large",
    help="use large download commands (only available from VivaTPS)",
    is_flag=True
)
def cli_download(**kwargs: Any) -> None:
    """Download a file from the instrument."""
    from .app import main_download

    main_download(**kwargs)



import os

from click_extra import (
    Choice,
    prompt,
    confirm,
    echo
)
from geocompy.communication import open_serial
from geocompy.geo import GeoCom
from geocompy.geo.gcdata import Prism

from ..utils import (
    echo_red,
    echo_green,
    echo_yellow
)
from ..targets import (
    TargetList,
    TargetPoint,
    load_targets_from_json,
    export_targets_to_json,
    import_targets_from_csv
)


def measure_targets(tps: GeoCom, filepath: str) -> TargetList | None:
    if os.path.exists(filepath):
        action: str = prompt(
            f"{filepath} already exists. Action",
            default="replace",
            type=Choice(["cancel", "replace", "append"])
        )
        match action:
            case "cancel":
                exit(0)
            case "append":
                points = load_targets_from_json(filepath)
                echo(f"Loaded targets: {points.get_target_names()}")
            case _:
                points = TargetList()
    else:
        points = TargetList()

    ptid: str
    while ptid := prompt("Point ID (or nothing to finish)", type=str):
        if ptid in points:
            remove = confirm(
                f"{ptid} already exists. Overwrite?"
            )
            if remove:
                points.pop_target(ptid)
            else:
                continue

        resp_target = tps.bap.get_prism_type()
        if resp_target.params is None:
            echo_yellow("Could not retrieve target type.")
            continue

        target = resp_target.params
        if target == Prism.USER:
            echo_yellow(
                "User defined prism types are currently not supported."
            )
            continue

        user_target: str = prompt(
            "Prism type",
            default=target.name,
            type=Choice([e.name for e in Prism if e.name != 'USER'])
        )
        target = Prism[user_target]

        resp_height = tps.tmc.get_target_height()
        if resp_height.params is None:
            echo_yellow("Could not retrieve target height.")
            continue

        height: float = prompt(
            "Target height",
            default=f"{resp_height.params:.4f}",
            type=float
        )

        prompt("Aim at target, then press ENTER...", prompt_suffix="")

        tps.aut.fine_adjust(0.5, 0.5)
        tps.tmc.do_measurement()
        resp = tps.tmc.get_simple_coordinate(10)
        if resp.params is None:
            echo_yellow("Could not measure target.")
            continue

        points.add_target(
            TargetPoint(
                ptid,
                target,
                height,
                resp.params
            )
        )

        echo(f"{ptid} stored")
        if not confirm("Record more targets?", default=True):
            break

    echo_green("Set measurement setup finished")

    return points


def main_measure(
    port: str,
    output: str,
    baud: int = 9600,
    timeout: int = 15,
    retry: int = 1,
    sync_after_timeout: bool = False
) -> None:

    with open_serial(
        port,
        retry=retry,
        sync_after_timeout=sync_after_timeout,
        speed=baud,
        timeout=timeout
    ) as com:
        tps = GeoCom(com)
        targets = measure_targets(tps, output)
        if targets is None:
            echo_red("Setup was cancelled or no targets were recorded.")
            exit(0)

    export_targets_to_json(output, targets)
    echo_green(f"Saved setup results at '{output}'")


def main_import(
    reflector: str,
    input: str,
    output: str,
    delimiter: str = ",",
    columns: str = "PENZ",
    skip: int = 0
) -> None:

    if os.path.exists(output):
        action: str = prompt(
            f"{output} already exists. Action",
            type=Choice(["cancel", "replace", "append"]),
            default="cancel"
        )
        match action:
            case "cancel":
                exit(0)
            case "append":
                points = load_targets_from_json(output)
                echo(
                    f"Loaded targets: {', '.join(points.get_target_names())}"
                )
            case _:
                points = TargetList()
    else:
        points = TargetList()

    try:
        imported_points = import_targets_from_csv(
            input,
            delimiter,
            columns,
            Prism[reflector],
            skip
        )
    except FileNotFoundError as fe:
        echo_red("Could not find CSV file (file does not exist)")
        echo_red(fe)
        exit(1103)
    except OSError as oe:
        echo_red(
            "Cannot import CSV data due to a file operation error "
            "(no access or other error)"
        )
        echo_red(oe)
        exit(1102)
    except Exception as e:
        echo_red(
            "Cannot import CSV data due to an error "
            "(duplicated points, the header was not skipped, malformed data "
            "or incorrect column spec)"
        )
        echo_red(e)
        exit(1100)

    conflicts = set(
        points.get_target_names()
    ).intersection(imported_points.get_target_names())

    if len(conflicts) > 0:
        echo(f"Duplicates: {', '.join(sorted(list(conflicts)))}")
        echo_red("Found duplicate targets between CSV and existing JSON")
        exit(1101)

    echo(f"Imported targets: {', '.join(imported_points.get_target_names())}",)

    for t in imported_points:
        points.add_target(t)

    export_targets_to_json(output, points)
    echo_green(f"Saved import results at '{os.path.abspath(output)}'")



from io import BufferedWriter

from click_extra import echo, progressbar
from geocompy.communication import open_serial
from geocompy.geo import GeoCom
from geocompy.geo.gctypes import GeoComCode
from geocompy.geo.gcdata import File, Device

from ..utils import echo_red, echo_green, echo_yellow


_FILE = {
    "image": File.IMAGE,
    "database": File.DATABASE,
    "overview-jpg": File.IMAGES_OVERVIEW_JPG,
    "overview-bmp": File.IMAGES_OVERVIEW_BMP,
    "telescope-jpg": File.IMAGES_TELESCOPIC_JPG,
    "telescope-bmp": File.IMAGES_TELESCOPIC_BMP,
    "scan": File.SCANS,
    "unknown": File.UNKNOWN,
    "last": File.LAST
}


_DEVICE = {
    "internal": Device.INTERNAL,
    "cf": Device.CFCARD,
    "sd": Device.SDCARD,
    "usb": Device.USB,
    "ram": Device.RAM
}


def run_listing(
    tps: GeoCom,
    dev: str,
    directory: str,
    filetype: str
) -> None:
    resp_setup = tps.ftr.setup_listing(
        _DEVICE[dev],
        _FILE[filetype],
        directory
    )
    if resp_setup.error != GeoComCode.OK:
        echo_red(f"Could not set up file listing ({resp_setup.error.name})")
        return

    resp_list = tps.ftr.list()
    if resp_list.error != GeoComCode.OK or resp_list.params is None:
        echo_red(f"Could not start listing ({resp_list.error.name})")
        return

    last, name, size, lastmodified = resp_list.params
    if name == "":
        echo_yellow("Directory is empty or path does not exist")
        return

    count = 1
    echo(f"{'file name':<55.55s}{'bytes':>10.10s}{'last modified':>25.25s}")
    echo(f"{'---------':<55.55s}{'-----':>10.10s}{'-------------':>25.25s}")
    fmt = "{name:<55.55s}{size:>10s}{date:>25.25s}"
    echo(
        fmt.format_map(
            {
                "name": name,
                "size": str(size),
                "date": (
                    lastmodified.isoformat(sep=" ")
                    if lastmodified is not None
                    else ""
                )
            }
        )
    )
    while not last:
        resp_list = tps.ftr.list(True)
        if resp_list.error != GeoComCode.OK or resp_list.params is None:
            echo_red(
                f"An error occured during listing ({resp_list.error.name})"
            )
            return

        last, name, size, lastmodified = resp_list.params
        echo(
            fmt.format_map(
                {
                    "name": name,
                    "size": str(size),
                    "date": (
                        lastmodified.isoformat(sep=" ")
                        if lastmodified is not None
                        else ""
                    )
                }
            )
        )
        count += 1

    echo("-" * 90)
    echo(f"total: {count} files")


def run_download(
    tps: GeoCom,
    filename: str,
    file: BufferedWriter,
    device: str = "internal",
    filetype: str = "unknown",
    chunk: int = 450,
    large: bool = False
) -> None:
    setup = tps.ftr.setup_download
    download = tps.ftr.download
    if large:
        setup = tps.ftr.setup_large_download
        download = tps.ftr.download_large

    resp_setup = setup(
        filename,
        chunk,
        _DEVICE[device],
        _FILE[filetype]
    )
    if resp_setup.error != GeoComCode.OK or resp_setup.params is None:
        echo_red(f"Could not set up file download ({resp_setup.error.name})")
        return

    block_count = resp_setup.params
    with progressbar(
        range(block_count),
        label="Downloading"
    ) as bar:
        for i in bar:
            resp_pull = download(i + 1)
            if resp_pull.error != GeoComCode.OK or resp_pull.params is None:
                echo_red(
                    "An error occured during download "
                    f"({resp_setup.error.name})"
                )
                return

            echo(bytes.fromhex(resp_pull.params), file, False)

    echo_green("Download complete")


def main_download(
    port: str,
    filename: str,
    output: BufferedWriter,
    baud: int = 9600,
    timeout: int = 15,
    retry: int = 1,
    sync_after_timeout: bool = False,
    device: str = "internal",
    filetype: str = "unknown",
    chunk: int = 450,
    large: bool = False
) -> None:
    with open_serial(
        port=port,
        speed=baud,
        timeout=timeout,
        retry=retry,
        sync_after_timeout=sync_after_timeout
    ) as com:
        tps = GeoCom(com)
        try:
            run_download(tps, filename, output, device, filetype, chunk, large)
        finally:
            tps.ftr.abort_download()


def main_list(
    port: str,
    directory: str = "/",
    baud: int = 9600,
    timeout: int = 15,
    retry: int = 1,
    sync_after_timeout: bool = False,
    device: str = "internal",
    filetype: str = "unknown"
) -> None:
    with open_serial(
        port=port,
        speed=baud,
        timeout=timeout,
        retry=retry,
        sync_after_timeout=sync_after_timeout
    ) as com:
        tps = GeoCom(com)
        try:
            run_listing(tps, device, directory, filetype)
        finally:
            tps.ftr.abort_list()



from typing import Any

from click_extra import (
    extra_command,
    argument,
    option,
    IntRange,
    File
)

from ..utils import (
    com_option_group,
    com_port_argument
)


@extra_command(
    "inclination",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@com_option_group()
@option(
    "-o",
    "--output",
    help="file to save output to",
    type=File("wt", encoding="utf8", lazy=True)
)
@option(
    "-p",
    "--positions",
    help="number of positions to measure around the circle",
    type=IntRange(1, 12),
    default=1
)
@option(
    "-z",
    "--zero",
    help="start from hz==0 (otherwise start from current orientation)",
    is_flag=True
)
@option(
    "-c",
    "--cycles",
    help="repetition cycles",
    type=IntRange(1),
    default=1
)
def cli_measure(**kwargs: Any) -> None:
    """Measure instrument inclination in multiple positions."""
    from .app import main_measure

    main_measure(**kwargs)


@extra_command(
    "inclination",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@argument(
    "input",
    help="inclination measurement file to process",
    type=File("rt", encoding="utf8")
)
@option(
    "-o",
    "--output",
    help="file to save results to in CSV format",
    type=File("wt", encoding="utf8", lazy=True)
)
def cli_calc(**kwargs: Any) -> None:
    """Calculate inclination from multiple measurements."""
    from .app import main_calc

    main_calc(**kwargs)


@extra_command(
    "inclination",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@argument(
    "output",
    help="output file",
    type=File("wt", encoding="utf8", lazy=True)
)
@argument(
    "inputs",
    help="inclination measurement files",
    type=File("rt", encoding="utf8"),
    nargs=-1,
    required=True
)
def cli_merge(**kwargs: Any) -> None:
    """Merge results from multiple inclination measurements."""
    from .app import main_merge

    main_merge(**kwargs)



from typing import Any

from click_extra import (
    extra_command,
    argument,
    option,
    Choice,
    IntRange,
    file_path,
    dir_path
)

from ..utils import (
    com_option_group,
    logging_option_group,
    com_port_argument
)


@extra_command(
    "sets",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@argument(
    "targets",
    type=file_path(exists=True),
    help="JSON file containing target definitions"
)
@argument(
    "directory",
    type=dir_path(),
    help="directory to save measurement output to"
)
@com_option_group()
@option(
    "-f",
    "--format",
    type=str,
    default="setmeasurement_{time}.json",
    help=(
        "session output file name format with placeholders "
        "(`{time}`: timestamp, `{order}`: order, `{cycle}`: cycles)"
    )
)
@option(
    "-c",
    "--cycles",
    type=IntRange(min=1),
    default=1,
    help="number of measurement cycles"
)
@option(
    "-o",
    "--order",
    help="measurement order (capital letter: face 1, lower case: face 2)",
    type=Choice(["AaBb", "AabB", "ABab", "ABba", "ABCD"]),
    default="ABba"
)
@option(
    "-s",
    "--sync-time",
    help="synchronize instrument time and date with the computer",
    is_flag=True
)
@option(
    "-p",
    "--points",
    type=str,
    help=(
        "targets to use from loaded target definition "
        "(comma separated list, empty to use all)"
    ),
    default=""
)
@logging_option_group()
def cli_measure(**kwargs: Any) -> None:
    """Run sets of measurements to predefined targets."""
    from .measure import main

    main(**kwargs)


@extra_command(
    "sets",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@argument(
    "output",
    help="output file",
    type=file_path()
)
@argument(
    "inputs",
    help="set measurement session JSON files (glob notation)",
    type=file_path(exists=True),
    nargs=-1,
    required=True
)
@option(
    "--allow-oneface",
    help="accept points with face 1 measurements only as well",
    is_flag=True
)
def cli_merge(**kwargs: Any) -> None:
    """Merge the output of multiple set measurement sessions.

    The results of every set measurement session are saved to a separate file.
    When multiple sessions are measured using the same targets from the same
    station, the data files need to be merged to process them together.

    .. note::
        The merge will be refused if the station information, or the target
        points do not match between the targeted sessions.
    """
    from .process import main_merge

    main_merge(**kwargs)


@extra_command(
    "sets",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@argument(
    "inputs",
    help="set measurement session JSON files (glob notation)",
    nargs=-1,
    required=True,
    type=file_path(exists=True)
)
@option(
    "-s",
    "--schema-only",
    help="only validate the JSON schema",
    is_flag=True
)
@option(
    "--allow-oneface",
    help="accept points with face 1 measurements only as well",
    is_flag=True
)
def cli_validate(**kwargs: Any) -> None:
    """Validate session output files.

    After the measurement sessions are finished, it might be useful to
    validate, that each session succeeded, no points were skipped.
    """
    from .process import main_validate

    main_validate(**kwargs)


@extra_command(
    "sets",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@argument(
    "input",
    help="input session file to process",
    type=file_path(exists=True)
)
@argument(
    "output",
    help="output CSV file",
    type=file_path(readable=False)
)
@option(
    "--header",
    help="write column headers",
    is_flag=True
)
@option(
    "-d",
    "--delimiter",
    help="column delimiter character",
    type=str,
    default=","
)
@option(
    "-p",
    "--precision",
    help="decimal precision",
    type=IntRange(min=0),
    default=4
)
@option(
    "--allow-oneface",
    help="accept points with face 1 measurements only as well",
    is_flag=True
)
def cli_calc(**kwargs: Any) -> None:
    """Calculate results from set measurements.

    The most common calculation needed after set measurements is the
    determination of the target coordinates, from results of multiple
    measurement sessions and/or cycles. The resulting coordinates (as well as
    their deviations) are saved to a simple CSV file.
    """
    from .process import main_calc

    main_calc(**kwargs)



from __future__ import annotations

import re
from enum import IntEnum
import logging
from typing import Any, cast, Callable
import inspect
from textwrap import dedent

from textual import on
from textual.suggester import Suggester
from textual.binding import Binding
from textual.app import App, ComposeResult
from textual.widgets import (
    Footer,
    Header,
    TabbedContent,
    TabPane,
    Button,
    Input,
    Label,
    Select,
    Log,
    TextArea
)
from textual.containers import Grid, HorizontalGroup
from textual.validation import Validator, ValidationResult
from rapidfuzz import fuzz, process
from geocompy.data import Angle, Coordinate
from geocompy.geo import GeoCom
from geocompy.gsi.dna import GsiOnlineDNA
from geocompy.communication import Connection, open_serial


class DummyGeoComConnection(Connection):
    class Port:
        def __init__(self, port: str):
            self.port = port

    _RESP = re.compile(
        r"^%R1P,"
        r"(?P<comrc>\d+),"
        r"(?P<tr>\d+):"
        r"(?P<rc>\d+)"
        r"(?:,(?P<params>.*))?$"
    )

    _CMD = re.compile(
        r"^%R1Q,"
        r"(?P<rpc>\d+)"
        r"(?P<trid>,\d+)?:"
        r"(?:(?P<params>.*))?$"
    )

    def __init__(self, port: str):
        self._port = self.Port(port)

    def __enter__(self) -> DummyGeoComConnection:
        return self

    def __exit__(self, *args: Any) -> None:
        return

    def close(self) -> None:
        return

    def send(self, message: str) -> None:
        return

    def receive(self) -> str:
        return ""

    def is_open(self) -> bool:
        return True

    def reset(self) -> None:
        return

    def exchange(self, cmd: str) -> str:
        if not self._CMD.match(cmd):
            return "%R1P,0,0:2"

        head, _ = cmd.split(":")
        match head.split(","):
            case [_, _, trid_str]:
                pass
            case _:
                trid_str = "0"

        trid = int(trid_str)

        return f"%R1P,0,{trid}:0"


# def open_serial(port: str, **kwargs: Any) -> DummyGeoComConnection:
#     return DummyGeoComConnection(port)


def get_app_logger(app: App[None], name: str) -> logging.Logger:
    log = logging.Logger(name, logging.DEBUG)
    fmt = logging.Formatter(
        "%(asctime)s <%(name)s> [%(levelname)s] %(message)s",
        "%Y-%m-%d %H:%M:%S"
    )
    handler = logging.StreamHandler(app.query_one("#log_com", Log))
    handler.setFormatter(fmt)
    log.addHandler(handler)
    return log


class Protocol(IntEnum):
    GEOCOM = 0
    GSIDNA = 1


_BAUD = [
    1200,
    2400,
    4800,
    9600,
    19200,
    38400,
    56000,
    57600,
    115200,
    230400,
    921600
]


class CmdSuggester(Suggester):
    PATH = re.compile(r"^(\w+(?:\.\w+)*)([\.\(]?)")

    def __init__(self, app: GeoComTerminal):
        super().__init__(use_cache=False, case_sensitive=False)
        self.app = app

    def get_last_valid_obj(self, value: str) -> tuple[Any, str, str] | None:
        if value == "":
            return self.app.protocol, "", "."

        mat = self.PATH.match(value)
        if mat is None:
            return None

        obj = self.app.protocol
        parts: list[str] = []
        namespaces = mat.group(1).split(".")
        for i in range(len(namespaces)):
            p = namespaces.pop(0)
            if not hasattr(obj, p):
                namespaces.insert(0, p)
                break

            obj = getattr(obj, p)
            parts.append(p)

        match len(namespaces):
            case 0:
                partial = "." if mat.group(2) == "." else ""
            case 1:
                partial = namespaces[0]
            case _:
                return None

        path = ".".join(parts)
        return obj, path, partial

    def format_signature(
        self,
        obj: Callable[[Any], Any],
        signature: inspect.Signature
    ) -> str:
        text = f"def {obj.__name__}("

        params: list[str] = []
        for param in signature.parameters.values():
            default = ""
            if param.default is not inspect._empty:
                default = f" = {param.default}"
            params.append(
                f"\t{param.name}: {param.annotation}{default}")

        if len(params) > 0:
            text += "\n" + ",\n".join(params) + "\n"

        text += f") -> {signature.return_annotation}\n"

        text += dedent(obj.__doc__) if obj.__doc__ is not None else ""

        return text

    async def get_suggestion(self, value: str) -> str | None:
        info = self.app.query_one("#cmd_info", TextArea)
        # log = self.app.query_one("#log_cmd", Log)

        mat = self.PATH.match(value)
        if value != "" and mat is None:
            # log.write_line("No match")
            info.text = ""
            return None

        result = self.get_last_valid_obj(value)
        if result is None:
            # log.write_line("No object")
            info.text = ""
            return None

        obj, path, partial = result
        if partial == "":
            # log.write_line("No partial")
            if value.removeprefix(path).startswith("("):
                try:
                    info.text = self.format_signature(
                        obj, inspect.signature(obj))
                except Exception:
                    info.text = ""
                return None

            info.text = ""
            return None

        suggest: list[str] = []
        if partial == ".":
            suggest = list(
                filter(lambda s: not s.startswith("_"), dir(obj)))
        else:
            suggest = list(
                filter(
                    lambda s: not s.startswith("_")
                    and fuzz.partial_ratio(partial, s) > 70,
                    dir(obj)
                )
            )

        # log.write_line(f"{suggest}, {obj}, {path}, {partial}")
        if len(suggest) == 0:
            info.text = ""
            return None

        suggestions = ", ".join(suggest)
        info.text = suggestions

        separator = "." if path != "" else ""

        suggestion = process.extractOne(
            partial, suggest, scorer=fuzz.partial_ratio)[0]

        return f"{path}{separator}{suggestion}"


class CmdInput(Input):
    BINDINGS = [
        Binding("escape", "clear", "Clear"),
        Binding("enter", "execute", "Execute"),
        Binding("up", "history('up')", "History"),
        Binding("down", "history('down')", "History")
    ]

    history: list[str] = []
    history_index: int | None = None

    PAT = re.compile(r"(\w+(?:\.\w+)*)\((.*)\)")

    def action_history(self, direction: str) -> None:
        if len(self.history) == 0:
            return

        match direction:
            case 'up':
                if self.history_index is None:
                    self.history_index = 0

                self.history_index = max(
                    self.history_index - 1, -len(self.history))
            case 'down':
                if self.history_index is None:
                    return

                self.history_index = min(self.history_index + 1, 0)
            case _:
                return

        self.value = self.history[self.history_index]
        self.cursor_position = len(self.value)

    def action_clear(self) -> None:
        self.history_index = None
        self.app.query_one("#cmd_info", TextArea).text = ""
        self.clear()

    def action_execute(self) -> None:
        if self.parent is None:
            return

        if not self.PAT.match(self.value):
            self.notify("Not a valid command!",
                        severity="error", title="Error")
            self.app.bell()
            return

        if len(self.history) == 0 or self.value != self.history[-1]:
            self.history.append(self.value)

        self.history_index = None
        app = cast(GeoComTerminal, self.app)
        app.action_execute()


class GeoComTerminal(App[None]):
    CSS_PATH = "app.tcss"
    TITLE = "I-man Interactive Terminal"

    protocol: GeoCom | GsiOnlineDNA | None = None

    def compose(self) -> ComposeResult:
        self.action_show_help_panel()

        with TabbedContent(initial="tab_com"):
            with TabPane("Connection", id="tab_com"):
                with Grid(classes="connect"):
                    yield Label("Port")
                    yield Input(
                        placeholder="COM1",
                        valid_empty=False,
                        validators=[
                            ComPort()
                        ],
                        id="edit_com"
                    )
                    yield Label("Protocol")
                    yield Select(
                        [
                            ("GeoCom", Protocol.GEOCOM),
                            ("GSI Online DNA", Protocol.GSIDNA)
                        ],
                        allow_blank=False,
                        id="select_protocol"
                    )
                    yield Label("Baud")
                    yield Select(
                        [(str(b), b) for b in _BAUD],
                        allow_blank=False,
                        value=9600,
                        id="select_baud"
                    )
                    yield Label("Timeout")
                    yield Input(
                        valid_empty=False,
                        value="15",
                        type="integer",
                        validators=[Timeout()],
                        id="edit_timeout"
                    )
                    yield Button(
                        "Update",
                        id="btn_update_timeout",
                        disabled=True
                    )
                with HorizontalGroup(id="hg_buttons"):
                    yield Button("Test Connection", id="btn_test_com")
                    yield Button(
                        "Connect",
                        id="btn_connect",
                        variant="primary"
                    )
                    yield Button(
                        "Disconnect",
                        id="btn_disconnect",
                        disabled=True,
                        variant="error"
                    )

            with TabPane("Commands", id="tab_cmd", disabled=True):
                yield Log(id="log_cmd", highlight=True)
                yield TextArea(id="cmd_info", read_only=True)
                yield CmdInput(id="edit_cmd", placeholder="Type commands...",
                               suggester=CmdSuggester(self),
                               select_on_focus=False)

            with TabPane("Logs", id="tab_logs"):
                yield Log(id="log_com", highlight=True)

        yield Header()
        yield Footer()

    def action_execute(self) -> None:
        cmd_input = self.query_one("#edit_cmd", CmdInput)
        cmd = cmd_input.value
        if cmd == "":
            return

        log = self.query_one("#log_cmd", Log)
        log.write_line(f">>> {cmd}")
        try:
            ans = eval(
                f"self.protocol.{cmd}",
                {},
                {
                    "self": self,
                    "Angle": Angle,
                    "Coordinate": Coordinate
                }
            )
            log.write_line(str(ans))
        except Exception as e:
            log.write_line(str(e))

        cmd_input.action_clear()
        self.query_one("#cmd_info", TextArea).text = ""

    def on_unmount(self) -> None:
        if self.protocol is None:
            return

        self.protocol._conn.close()
        self.protocol = None

    @on(Button.Pressed, "#btn_test_com")
    def btn_test_com_pressed(self, event: Button.Pressed) -> None:
        if event.button is None:
            return

        port = self.query_one("#edit_com", Input)
        if not port.is_valid:
            self.notify(
                "Invalid port identifier given!",
                severity="error",
                title="Error"
            )
            self.bell()
            return
        baud = cast(int, self.query_one("#select_baud", Select).value)
        timeout = self.query_one("#edit_timeout", Input)
        if not timeout.is_valid:
            self.notify(
                "Invalid timeout value given!",
                severity="error",
                title="Error"
            )
            self.app.bell()
            return
        try:
            with open_serial(
                port.value,
                speed=baud,
                timeout=int(timeout.value)
            ) as com:
                match self.query_one("#select_protocol", Select).value:
                    case Protocol.GEOCOM:
                        ans = com.exchange(r"%R1Q,0:")
                        if ans != "%R1P,0,0:0":
                            raise Exception("Invalid response.")
                    case Protocol.GSIDNA:
                        ans = com.exchange(r"a")
                        if ans != "?":
                            raise Exception("Invalid response.")

        except Exception as e:
            self.notify(
                f"Connection test failed. ({e})",
                severity="error",
                title="Error"
            )
            self.bell()
            return

        self.notify("Connection test successful!", title="Success")
        self.bell()
        return

    @on(Button.Pressed, "#btn_connect")
    def btn_connect_pressed(self, event: Button.Pressed) -> None:
        if event.button is None:
            return
        port = self.query_one("#edit_com", Input)
        if not port.is_valid:
            self.notify(
                "Invalid port identifier given!",
                severity="error",
                title="Error"
            )
            self.app.bell()
            return
        baud = cast(int, self.query_one("#select_baud", Select).value)
        timeout = self.query_one("#edit_timeout", Input)
        if not timeout.is_valid:
            self.notify(
                "Invalid timeout value given!",
                severity="error",
                title="Error"
            )
            self.app.bell()
            return
        try:
            log = get_app_logger(self, port.value)
            com = open_serial(
                port.value,
                speed=baud,
                timeout=int(timeout.value)
            )
            match self.query_one("#select_protocol", Select).value:
                case Protocol.GEOCOM:
                    self.protocol = GeoCom(com, log)
                    self.sub_title = f"GeoCom ({com._port.port})"
                case Protocol.GSIDNA:
                    self.protocol = GsiOnlineDNA(com, log)
                    self.sub_title = f"GSI Online DNA ({com._port.port})"

            self.query_one("#btn_test_com").disabled = True
            self.query_one("#btn_disconnect").disabled = False
            event.button.disabled = True
            self.query_one("#edit_com", Input).disabled = True
            self.query_one("#select_protocol", Select).disabled = True
            self.query_one("#select_baud", Select).disabled = True
            self.query_one("#btn_update_timeout", Button).disabled = False
            self.query_one("#tab_cmd", TabPane).disabled = False

            self.notify("Connection successful.", title="Success")
            self.bell()

        except Exception as e:
            self.notify(
                f"Could not connect. ({e})",
                severity="error",
                title="Error"
            )
            self.bell()

    @on(Button.Pressed, "#btn_disconnect")
    def btn_disconnect_pressed(self, event: Button.Pressed) -> None:
        if event.button is None:
            return

        if self.protocol is None:
            return

        try:
            self.protocol._conn.close()
            self.protocol = None
            self.query_one("#btn_test_com").disabled = False
            self.query_one("#btn_connect").disabled = False
            event.button.disabled = True
            self.query_one("#edit_com", Input).disabled = False
            self.query_one("#select_protocol", Select).disabled = False
            self.query_one("#select_baud", Select).disabled = False
            self.query_one("#btn_update_timeout", Button).disabled = True
            self.query_one("#tab_cmd", TabPane).disabled = True
            self.sub_title = ""
        except Exception as e:
            self.notify(
                f"Could not disconnect. ({e})",
                severity="error",
                title="Error"
            )
            self.bell()
            return

        self.notify("Disconnected.", title="Success")
        self.bell()

    @on(Button.Pressed, "#btn_update_timeout")
    def btn_update_timeout_pressed(self, event: Button.Pressed) -> None:
        if self.protocol is None:
            return

        edit = self.query_one("#edit_timeout", Input)
        if not edit.is_valid:
            self.notify(
                f"{edit.value} is not a valid timeout",
                title="Error",
                severity="error"
            )
            return

        com = self.protocol._conn
        try:
            com._port.timeout = int(  # type: ignore[attr-defined]
                edit.value
            )
            self.notify(
                "Timeout updated",
                title="Success",
                severity="information"
            )
        except Exception:
            self.notify(
                f"{edit.value} is not a valid timeout",
                title="Error",
                severity="error"
            )


class ComPort(Validator):
    _WINCOM = re.compile(r"COM\d+")
    _LINUXCOM = re.compile(r"/dev/\w+")

    def validate(self, value: str) -> ValidationResult:
        if self._WINCOM.match(value) or self._LINUXCOM.match(value):
            return self.success()

        return self.failure("Not a valid COM port")


class Timeout(Validator):
    def validate(self, value: str) -> ValidationResult:
        if value != "-" and int(value) < 0:
            return self.failure("Timeout cannot be negative")

        return self.success()



from click_extra import extra_group, version_option

try:
    from ._version import __version__ as __version__
except Exception:
    __version__ = "0.0.0"  # Placeholder value for source installs

from . import morse
from . import terminal
from . import setup
from . import setmeasurement
from . import protocoltest
from . import inclination
from . import filetransfer
from . import jobs
from . import datatransfer
from . import settings


@extra_group("iman", params=None)  # type: ignore[misc]
@version_option()
def cli() -> None:
    """Automated measurement programs and related utilities for surveying
    instruments."""
    pass


@cli.group("measure")  # type: ignore[misc]
def cli_measure() -> None:
    """Conduct measurements."""


@cli.group("import")  # type: ignore[misc]
def cli_import() -> None:
    """Import external data and convert it for use with other commands."""


@cli.group("calc")  # type: ignore[misc]
def cli_calc() -> None:
    """Preform calculations from measurement results."""


@cli.group("merge")  # type: ignore[misc]
def cli_merge() -> None:
    """Merge various output files."""


@cli.group("validate")  # type: ignore[misc]
def cli_validate() -> None:
    """Validate intermediate files."""


@cli.group("test")  # type: ignore[misc]
def cli_test() -> None:
    """Test protocol responsiveness."""


@cli.group("list")  # type: ignore[misc]
def cli_list() -> None:
    """List various data stored on the instrument."""


@cli.group("download")  # type: ignore[misc]
def cli_download() -> None:
    """Download data from the instrument."""


@cli.group("upload")  # type: ignore[misc]
def cli_upload() -> None:
    """Upload data to the instrument."""


cli.add_command(morse.cli)
cli.add_command(terminal.cli)
cli_measure.add_command(setmeasurement.cli_measure)
cli_measure.add_command(setup.cli_measure)
cli_measure.add_command(inclination.cli_measure)
cli_calc.add_command(setmeasurement.cli_calc)
cli_calc.add_command(inclination.cli_calc)
cli_test.add_command(protocoltest.cli_geocom)
cli_test.add_command(protocoltest.cli_gsidna)
cli_merge.add_command(setmeasurement.cli_merge)
cli_merge.add_command(inclination.cli_merge)
cli_validate.add_command(setmeasurement.cli_validate)
cli_validate.add_command(settings.cli_validate)
cli_import.add_command(setup.cli_import)
cli_list.add_command(filetransfer.cli_list)
cli_list.add_command(jobs.cli_list)
cli_download.add_command(filetransfer.cli_download)
cli_download.add_command(datatransfer.cli_download)
cli_download.add_command(settings.cli_download)
cli_upload.add_command(datatransfer.cli_upload)
cli_upload.add_command(settings.cli_upload)



from click_extra import (
    echo,
    pause
)
from serial import SerialException
from geocompy.data import Angle
from geocompy.geo import GeoCom
from geocompy.geo.gctypes import GeoComCode
from geocompy.gsi.dna import GsiOnlineDNA
from geocompy.communication import open_serial

from ..utils import (
    echo_red,
    echo_green,
    echo_yellow
)


def tests_geocom(tps: GeoCom) -> None:
    echo("GeoCom connection successful")
    echo(
        "Various GeoCom functions will be tested. Certain settings will be "
        "changed on the instrument (ATR off, prism target off, etc.)."
    )
    echo(
        "The program will attempt to use motorized functions. Give "
        "appropriate clearance for the instrument!"
    )
    pause("Press any key when ready to proceed...")

    echo("Testing subsystems...")
    resp_aus = tps.aus.switch_user_atr(False)
    if resp_aus.error == GeoComCode.OK:
        echo_green("Alt User available")
    else:
        echo_yellow(f"Alt User unavailable ({resp_aus.response})")

    tps.aut.switch_atr(False)
    resp_aut = tps.aut.turn_to(0, Angle(180, 'deg'))
    if resp_aut.error == GeoComCode.OK:
        echo_green("Automation available")
    else:
        echo_yellow(f"Automation unavailable ({resp_aut.response})")

    resp_bap = tps.bap.get_measurement_program()
    if resp_bap.error == GeoComCode.OK:
        echo_green("Basic Applications available")
    else:
        echo_yellow(f"Basic Applications unavailable ({resp_bap.response})")

    resp_bmm = tps.bmm.beep_normal()
    if resp_bmm.error == GeoComCode.OK:
        echo_green("Basic Man-Machine interface available")
    else:
        echo_yellow(
            "Basic Man-Machine interface unavailable "
            f"({resp_bmm.response})"
        )

    resp_cam = tps.cam.set_focus_to_infinity()
    if resp_cam.error == GeoComCode.OK:
        echo_green("Camera available")
    else:
        echo_yellow(f"Camera unavailable ({resp_cam.response})")

    echo_green("Communcation available")

    resp_csv = tps.csv.get_instrument_name()
    if resp_csv.error == GeoComCode.OK:
        echo_green("Central Services available")
    else:
        echo_yellow(f"Central Services unavailable ({resp_csv.response})")

    resp_ctl = tps.ctl.get_wakeup_counter()
    if resp_ctl.error == GeoComCode.OK:
        echo_green("Control Task available")
    else:
        echo_yellow(f"Control Task unavailable ({resp_ctl.response})")

    resp_dna = tps.dna.switch_staffmode(False)
    if resp_dna.error == GeoComCode.OK:
        echo_green("Digital Level available")
    else:
        echo_yellow(f"Digital Level unavailable ({resp_dna.response})")

    resp_edm = tps.edm.switch_laserpointer(False)
    if resp_edm.error == GeoComCode.OK:
        echo_green("Electronic Distance Measurement available")
    else:
        echo_yellow(
            "Electronic Distance Measurement unavailable "
            f"({resp_edm.response})"
        )

    resp_ftr = tps.ftr.setup_listing()
    if resp_ftr.error == GeoComCode.OK:
        echo_green("File Transfer available")
        tps.ftr.abort_list()
    else:
        echo_yellow(f"File Transfer unavailable ({resp_ftr.response})")

    resp_img = tps.img.get_telescopic_configuration()
    if resp_img.error == GeoComCode.OK:
        echo_green("Imaging available")
    else:
        echo_yellow(f"Imaging unavailable ({resp_img.response})")

    resp_kdm = tps.kdm.get_display_power_status()
    if resp_kdm.error == GeoComCode.OK:
        echo_green("Keyboard Display Unit available")
    else:
        echo_yellow(f"Keyboard Display Unit unavailable ({resp_kdm.response})")

    tps.mot.stop_controller()
    resp_mot = tps.mot.start_controller()
    if resp_mot.error == GeoComCode.OK:
        echo_green("Motorization available")
    else:
        echo_yellow(f"Motorization unavailable ({resp_mot.response})")

    resp_sup = tps.sup.get_poweroff_configuration()
    if resp_sup.error == GeoComCode.OK:
        echo_green("Supervisor available")
    else:
        echo_yellow(f"Supervisor unavailable ({resp_sup.response})")

    resp_tmc = tps.tmc.get_station()
    if resp_tmc.error == GeoComCode.OK:
        echo_green("Theodolite Measurement and Calculation available")
    else:
        echo_yellow(
            "Theodolite Measurement and Calculation unavailable "
            f"({resp_tmc.response})"
        )

    resp_wir = tps.wir.get_recording_format()
    if resp_wir.error == GeoComCode.OK:
        echo_green("Word Index Registration available")
    else:
        echo_yellow(
            "Word Index Registration unavailable "
            f"({resp_wir.response})"
        )


def tests_gsidna(dna: GsiOnlineDNA) -> None:
    echo("GSI Online connection successful")
    echo(
        "Various GSI Online DNA functions will be tested. Certain settings "
        "might be changed on the instrument (staff mode, point number, etc.)."
    )
    pause("Press any key when ready to proceed...")

    echo("Testing settings...")
    staff_get = dna.settings.get_staff_mode()
    if staff_get.value is None:
        echo_yellow(f"Settings queries unavailable ({staff_get.response})")
    else:
        echo_green("Settings queries available")

    staff_set = dna.settings.set_staff_mode(False)
    if not staff_set.value:
        echo_yellow(f"Settings commands unavailable ({staff_set.response})")
    else:
        echo_green("Settings commands available")

    echo("Testing measurements...")
    point_get = dna.measurements.get_point_id()
    if point_get.value is None:
        echo_yellow(
            f"Measurement/database queries unavailable ({point_get.response})"
        )
    else:
        echo_green("Measurement/database queries available")

    point_set = dna.measurements.set_point_id("TEST")
    if not point_set.value:
        echo_yellow(
            f"Measurement/database commands unavailable ({point_set.response})"
        )
    else:
        echo_green("Measurement/database commands available")


def main(
    port: str,
    protocol: str,
    baud: int = 9600,
    timeout: int = 15,
    retry: int = 1,
    sync_after_timeout: bool = False
) -> None:
    try:
        with open_serial(
            port,
            speed=baud,
            timeout=timeout,
            retry=retry,
            sync_after_timeout=sync_after_timeout
        ) as com:
            try:
                if protocol == "geocom":
                    tps = GeoCom(com)
                    tests_geocom(tps)
                elif protocol == "gsidna":
                    dna = GsiOnlineDNA(com)
                    tests_gsidna(dna)
            except Exception as e:
                echo_red("An exception occured while running the tests")
                echo_red(e)

    except (SerialException, ConnectionError) as e:
        echo_red(f"Connection was not successful ({e})")



import math


def adjust_uniform_single(values: list[float]) -> tuple[float, float]:
    n = len(values)
    adjusted = math.fsum(values) / n
    dev = math.sqrt(math.fsum([(v - adjusted)**2 for v in values]) / n)
    return adjusted, dev



from io import TextIOWrapper
from time import sleep
from math import tan, atan, degrees
from re import compile

from click_extra import echo, progressbar
from geocompy.data import Angle, Coordinate
from geocompy.geo import GeoCom
from geocompy.communication import open_serial

from ..calculations import adjust_uniform_single
from ..utils import echo_green


_LINE = compile(r"^\d+(?:\.\d+)?(?:,\-?\d+\.\d+){2}$")


def run_measure(
    tps: GeoCom,
    output: TextIOWrapper | None = None,
    positions: int = 1,
    zero: bool = False,
    cycles: int = 1
) -> None:
    turn = 360 // positions
    v = Angle(90, 'deg')
    start = 0

    if not zero:
        angles = tps.tmc.get_angle()
        if angles.params is not None:
            start = round(angles.params[0].asunit('deg'))

    echo(
        "hz_deg,cross_sec,length_sec",
        output
    )

    with progressbar(
        range(start, start + cycles * 360, turn),
        label="Measuring",
        hidden=output is None
    ) as bar:
        for a in bar:
            hz = Angle(a, 'deg').normalized()
            tps.aut.turn_to(hz, v)

            sleep(1)
            fullangles = tps.tmc.get_angle_inclination('MEASURE')
            if fullangles.params is None:
                continue

            az = fullangles.params[0]
            cross = fullangles.params[4]
            length = fullangles.params[5]

            echo(
                f"{az.asunit('deg'):.4f},"
                f"{cross.asunit('deg') * 3600:.2f},"
                f"{length.asunit('deg') * 3600:.2f}",
                output
            )


def main_measure(
    port: str,
    baud: int = 9600,
    timeout: int = 15,
    retry: int = 1,
    sync_after_timeout: bool = False,
    output: TextIOWrapper | None = None,
    positions: int = 1,
    zero: bool = False,
    cycles: int = 1
) -> None:
    with open_serial(
        port,
        retry=retry,
        sync_after_timeout=sync_after_timeout,
        speed=baud,
        timeout=timeout
    ) as com:
        tps = GeoCom(com)
        run_measure(tps, output, positions, zero, cycles)


def main_merge(
    inputs: list[TextIOWrapper],
    output: TextIOWrapper
) -> None:
    echo("hz_deg,cross_sec,length_sec", output)
    for item in inputs:
        for line in item:
            if not _LINE.match(line.strip()):
                continue

            echo(line, output, False)

    echo_green(f"Merged measurements from {len(inputs)} files.")


def main_calc(
    input: TextIOWrapper,
    output: TextIOWrapper | None = None
) -> None:
    points: list[Coordinate] = []

    for line in input:
        if not _LINE.match(line.strip()):
            continue

        fields = line.strip().split(",")
        azimut = Angle(float(fields[0]), 'deg')
        cross = Angle(float(fields[1]) / 3600, 'deg')
        length = Angle(float(fields[2]) / 3600, 'deg')

        coord = Coordinate(tan(cross), tan(length), 1).normalized()
        bearing, inclination, s = coord.to_polar()

        points.append(
            Coordinate.from_polar(
                (bearing + azimut).normalized(),
                inclination,
                s
            )
        )

    x, x_dev = adjust_uniform_single([p.x for p in points])
    y, y_dev = adjust_uniform_single([p.y for p in points])
    z, _ = adjust_uniform_single([p.z for p in points])

    inc_x = degrees(atan(x)) * 3600
    inc_y = degrees(atan(y)) * 3600
    inc_x_dev = degrees(atan(x_dev)) * 3600
    inc_y_dev = degrees(atan(y_dev)) * 3600

    direction, inc, _ = Coordinate(x, y, z).to_polar()

    if output is None:
        echo(f"""Axis aligned:
    inclination X: {inc_x:.1f}" +/- {inc_x_dev:.1f}"
    inclination Y: {inc_y:.1f}" +/- {inc_y_dev:.1f}"
Polar:
    direction: {direction.asunit('deg'):.4f}°
    inclination: {inc.asunit('deg') * 3600:.1f}\""""
             )
        return

    echo(
        "inc_x_sec,inc_x_dev_sec,inc_y_sec,inc_y_dev_sec,dir_deg,inc_sec",
        output
    )

    echo(
        (
            f"{inc_x:.1f},{inc_x_dev:.1f},{inc_y:.1f},{inc_y_dev:.1f},"
            f"{direction.asunit('deg'):.4f},{inc.asunit('deg') * 3600:.1f}"
        ),
        output
    )



from typing import Any

from click_extra import extra_command

from ..utils import (
    com_option_group,
    com_port_argument
)


@extra_command(
    "jobs",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@com_option_group()
def cli_list(**kwargs: Any) -> None:
    """List job files on an instrument."""
    from .app import main_list

    main_list(**kwargs)



from pathlib import Path
import json
import os

from jsonschema import validate, ValidationError
from click_extra import echo

from ..utils import echo_green, echo_red
from .io import read_settings, SettingsDict


def validate_settings(
    settings: SettingsDict,
    print_error: bool = False
) -> bool:
    with open(
        os.path.join(
            os.path.dirname(__file__),
            "schema_settings.json"
        ),
        "rt",
        encoding="utf8"
    ) as schema:
        try:
            validate(settings, json.load(schema))
        except ValidationError as e:
            if print_error:
                echo(e.message)
            return False

    return True


def main(
    file: Path,
    format: str = "auto"
) -> None:
    settings = read_settings(file, format)
    is_valid = validate_settings(settings, print_error=True)
    if not is_valid:
        echo_red("Settings file is not valid")
        return

    echo_green("Settings file is valid")



from io import BufferedWriter, TextIOWrapper

from serial import SerialTimeoutException
from click_extra import echo, progressbar
from geocompy.communication import open_serial

from ..utils import echo_green, echo_red, echo_yellow


def main_download(
    port: str,
    baud: int = 9600,
    timeout: int = 2,
    output: BufferedWriter | None = None,
    eof: str = "",
    autoclose: bool = True,
    include_eof: bool = False
) -> None:
    eof_bytes = eof.encode("ascii")
    with open_serial(
        port,
        speed=baud,
        timeout=timeout
    ) as com:
        eol_bytes = com.eombytes
        started = False
        while True:
            try:
                data = com.receive_binary()
                started = True

                if data == eof_bytes and autoclose and not include_eof:
                    echo_green("Download finished (end-of-file)")
                    return

                echo(data.decode("ascii", "replace"))
                if output is not None:
                    output.write(data + eol_bytes)

                if data == eof_bytes and autoclose:
                    echo_green("Download finished (end-of-file)")
                    return
            except SerialTimeoutException:
                if started and autoclose:
                    echo_green("Download finished (timeout)")
                    return
            except KeyboardInterrupt:
                echo_yellow("Download stopped manually")
                return
            except Exception as e:
                echo_red(f"Download interrupted by error ({e})")
                return


def main_upload(
    port: str,
    file: TextIOWrapper,
    baud: int = 1200,
    timeout: int = 15,
    skip: int = 0
) -> None:
    with open_serial(
        port,
        speed=baud,
        timeout=timeout
    ) as com:
        try:
            for _ in range(skip):
                next(file)

            count = 0
            with progressbar(
                file,
                label="Uploading data",
                item_show_func=lambda x: f"{count} line(s)"
            ) as bar:
                for line in bar:
                    com.send(line)
                    count += 1
        except Exception as e:
            echo_red(f"Upload interrupted by error ({e})")
            return

        echo_green("Upload finished")



from typing import Any

from click_extra import (
    extra_command,
    argument,
    option,
    IntRange,
    Choice
)

from ..utils import (
    com_option_group,
    com_port_argument
)


@extra_command(
    "targets",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@argument(
    "output",
    help=(
        "path to save the JSON containing the recorded targets "
        "(if the file already exists, the new targets can be appended)"
    ),
    type=str
)
@com_option_group()
def cli_measure(**kwargs: Any) -> None:
    """Measure target points.

    The program gives instructions in the terminal at each step.

    .. caution::
        :class: warning

        The appropriate prism type needs to be set on the instrument before
        recording each target point. The program will automatically request
        the type from the instrument after the point is measured.
    """
    from .app import main_measure

    main_measure(**kwargs)


@extra_command(
    "targets",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@argument(
    "reflector",
    help="prism type of the targets",
    type=Choice(
        (
            'ROUND',
            'MINI',
            'TAPE',
            'THREESIXTY',
            'USER1',
            'USER2',
            'USER3',
            'MINI360',
            'MINIZERO',
            'NDSTAPE',
            'GRZ121',
            'MPR122'
        )
    )
)
@argument(
    "input",
    help="csv file containing the target coordinates",
    type=str
)
@argument(
    "output",
    help="path to save the target definition to",
    type=str
)
@option(
    "-d",
    "--delimiter",
    help="column delimiter character",
    type=str,
    default=","
)
@option(
    "-c",
    "--columns",
    help=(
        "column spec "
        "(P: point ID, E: easting, N: northing, Z: height, _: ignore)"
    ),
    type=str,
    default="PENZ"
)
@option(
    "-s",
    "--skip",
    help="number of header rows to skip",
    type=IntRange(min=0),
    default=0
)
def cli_import(**kwargs: Any) -> None:
    """Import target points.

    If a coordinate list already exists with the target points, it can
    be imported from CSV format.

    As a CSV file may contain any number and types of columns, the
    mapping to the relevant columns can be given with a column spec.
    A column spec is a string, with each character representing a
    column type.

    - ``P``: point ID
    - ``E``: easting
    - ``N``: northing
    - ``Z``: up/height
    - ``_``: ignore/skip column

    Every column spec must specify the ``PENZ`` fields in the appropriate
    order.

    Examples:

    - ``PENZ``: standard column order
    - ``P_ENZ``: skipping 2nd column containing point codes
    - ``EN_Z_P``: mixed column order and skipping
    """
    from .app import main_import

    main_import(**kwargs)



import os
import json
import math
from typing import Sequence
import pathlib

from jmespath import search
from jsonschema import validate, ValidationError
from geocompy.data import Angle, Coordinate

from .sessions import SessionDict
from ..utils import (
    echo_red,
    echo_green,
    echo_yellow,
    make_directory
)


class SessionValidator:
    def __init__(
        self,
        twoface: bool
    ) -> None:
        self._twoface = twoface
        with open(
            os.path.join(
                os.path.dirname(__file__),
                "schema_session.json"
            ),
            "rt",
            encoding="utf8"
        ) as file_schema:
            self._schema = json.load(file_schema)

    def validate(self, data: SessionDict) -> None:
        validate(data, self._schema)

        ptids = search("cycles[*].points[*].name", data)
        if len(ptids) == 0:
            return

        first = ptids[0]
        for ids in ptids[1:]:
            if ids != first:
                raise ValueError("Mismatching points between cycles")

        if not self._twoface:
            return

        for i, c in enumerate(data["cycles"]):
            for p in c["points"]:
                if p.get("face2") is None:
                    raise ValueError(
                        f"Point {p['name']} is missing face 2 "
                        f"measurements in cycle {i + 1}"
                    )

    def validate_for_merge(self, data: Sequence[SessionDict]) -> None:
        for s in data:
            validate(s, self._schema)


def calc_angles(
    hz_f1: Angle,
    v_f1: Angle,
    hz_f2: Angle,
    v_f2: Angle
) -> tuple[Angle, Angle, Angle, Angle]:
    temp = hz_f2 - hz_f1
    if temp < 0:
        collim = (temp + Angle(180, 'deg')) / 2
    else:
        collim = (temp - Angle(180, 'deg')) / 2

    hz = hz_f1 + collim

    index = (Angle(360, 'deg') - v_f1 - v_f2) / 2
    v = v_f1 + index

    return hz, v, collim, index


def calc_coords(
    coords: list[Coordinate]
) -> tuple[Coordinate, Coordinate]:
    def adjust(values: list[float]) -> tuple[float, float]:
        n = len(values)
        adjusted = math.fsum(values) / n
        dev = math.sqrt(math.fsum([(v - adjusted)**2 for v in values]) / n)
        return adjusted, dev

    xi = [c.x for c in coords]
    yi = [c.y for c in coords]
    zi = [c.z for c in coords]

    x, x_dev = adjust(xi)
    y, y_dev = adjust(yi)
    z, z_dev = adjust(zi)

    return Coordinate(x, y, z), Coordinate(x_dev, y_dev, z_dev)


def main_merge(
    output: pathlib.Path,
    inputs: tuple[pathlib.Path],
    allow_oneface: bool = False
) -> None:
    sessions: list[SessionDict] = []
    for path in inputs:
        with path.open("rt", encoding="utf8") as file:
            sessions.append(json.load(file))

    validator = SessionValidator(not allow_oneface)
    try:
        validator.validate_for_merge(sessions)
    except ValidationError as ve:
        echo_red("One of the input files does not follow the required schema")
        echo_red(ve)
        exit(4)

    if len(sessions) == 0:
        echo_yellow("There were no sessions found to merge")
        exit(0)

    session: SessionDict = {
        "station": sessions[0]["station"],
        "instrumentheight": sessions[0]["instrumentheight"],
        "cycles": [c for s in sessions for c in s["cycles"]]
    }

    try:
        validator.validate(session)
    except ValidationError as ve:
        echo_red("The merging process caused a schema error")
        echo_red(ve)
        exit(4)
    except ValueError as e:
        echo_red(f"The merged output could not be validated ({e})")
        exit(4)

    make_directory(str(output))
    with open(output, "wt", encoding="utf8") as file:
        json.dump(
            session,
            file,
            indent=4
        )

    echo_green(
        f"Merged {len(session['cycles'])} cycles "
        f"from {len(sessions)} sessions"
    )


def main_validate(
    inputs: tuple[pathlib.Path],
    schema_only: bool = False,
    allow_oneface: bool = False
) -> None:
    sessions: list[SessionDict] = []
    for path in inputs:
        with path.open("rt", encoding="utf8") as file:
            sessions.append(json.load(file))

    validator = SessionValidator(not allow_oneface)
    if schema_only:
        try:
            validator.validate_for_merge(sessions)
            echo_green("Schema validation succeeded")
            exit(0)
        except ValidationError as ve:
            echo_red(
                "One of the input files does not follow the required schema"
            )
            echo_red(ve)
            exit(4)

    try:
        for s in sessions:
            validator.validate(s)

        echo_green("Validation succeeded")
        exit(0)
    except ValidationError as ve:
        echo_red(
            "One of the input files does not follow the required schema"
        )
        echo_red(ve)
        exit(4)
    except ValueError as e:
        echo_red(f"The merged output could not be validated ({e})")
        exit(4)


def main_calc(
    input: pathlib.Path,
    output: pathlib.Path,
    header: bool = False,
    delimiter: str = ",",
    precision: int = 4,
    allow_oneface: bool = False
) -> None:
    with input.open("rt", encoding="utf8") as file:
        data: SessionDict = json.load(file)

    validator = SessionValidator(not allow_oneface)
    try:
        validator.validate(data)
    except ValidationError as ve:
        echo_red("Input data does not follow the required schema")
        echo_red(ve)
        exit(4)
    except ValueError as e:
        echo_red("The input data did not pass validation")
        echo_red(e)
        exit(4)

    points = {"points": search("cycles[].points[]", data)}
    ptids = list(set(search("points[].name", points)))

    station = Coordinate(
        *data["station"]
    ) + Coordinate(
        0,
        0,
        data["instrumentheight"]
    )
    coords: dict[str, list[Coordinate]] = {}
    for pt in ptids:
        measurements = search(
            f"points[?name=='{pt}'].[height, face1, face2]",
            points
        )
        if pt not in coords:
            coords[pt] = []
        for cycle in measurements:
            height = cycle[0]
            hz = Angle(cycle[1][0])
            v = Angle(cycle[1][1])
            d = cycle[1][2]

            if cycle[2] is not None:
                hz_f2 = Angle(cycle[2][0])
                v_f2 = Angle(cycle[2][1])
                d_f2 = cycle[2][2]

                hz, v, co, z = calc_angles(
                    hz,
                    v,
                    hz_f2,
                    v_f2
                )

                d = (d + d_f2) / 2
            elif not allow_oneface:
                echo_red("Not all measurements have data for both faces")
                exit(4)

            c = station + Coordinate.from_polar(
                hz,
                v,
                d
            ) - Coordinate(
                0,
                0,
                height
            )
            coords[pt].append(c)

    final: list[tuple[str, Coordinate, Coordinate]] = []
    for name, coo in coords.items():
        coord, dev = calc_coords(coo)
        final.append((name, coord, dev))

    with output.open("wt", encoding="utf8") as file:
        if header:
            file.write(
                delimiter.join(
                    ["id", "e", "n", "h", "sigma_e", "sigma_n", "sigma_h"]
                ) + "\n"
            )

        fmt = "{0:." + str(precision) + "f}"
        for name, coord, dev in final:
            fields = [
                name,
                fmt.format(coord.x),
                fmt.format(coord.y),
                fmt.format(coord.z),
                fmt.format(dev.x),
                fmt.format(dev.y),
                fmt.format(dev.z)
            ]
            file.write(
                delimiter.join(fields) + "\n"
            )



from __future__ import annotations

import json
from datetime import datetime
from typing import TypedDict, NotRequired

from geocompy.data import Angle, Coordinate
from geocompy.geo.gcdata import Face

from ..utils import make_directory


class PointDict(TypedDict):
    name: str
    height: float
    face1: tuple[float, float, float]
    face2: NotRequired[tuple[float, float, float]]


class CycleDict(TypedDict):
    time: str
    battery: float | None
    inclination: tuple[float, float] | None
    temperature: float | None
    points: list[PointDict]


class SessionDict(TypedDict):
    station: tuple[float, float, float]
    instrumentheight: float
    cycles: list[CycleDict]


class Point:
    def __init__(
        self,
        name: str,
        height: float
    ) -> None:
        self.name = name
        self.height = height
        self.face1: tuple[Angle, Angle, float] | None = None
        self.face2: tuple[Angle, Angle, float] | None = None

    @classmethod
    def from_dict(cls, data: PointDict) -> Point:
        output = cls(
            data["name"],
            data["height"]
        )
        if data.get("face1") is not None:
            output.face1 = (
                Angle(data["face1"][0]),
                Angle(data["face1"][1]),
                data["face1"][2]
            )
        if data.get("face2") is not None:
            output.face2 = (
                Angle(data["face2"][0]),
                Angle(data["face2"][1]),
                data["face2"][2]
            )

        return output

    def to_dict(self) -> PointDict:
        if self.face1 is None:
            raise ValueError("Cannot export point due to missing F1 data")

        output: PointDict = {
            "name": self.name,
            "height": self.height,
            "face1": (
                float(self.face1[0]),
                float(self.face1[1]),
                self.face1[2]
            )
        }

        if self.face2 is not None:
            output["face2"] = (
                float(self.face2[0]),
                float(self.face2[1]),
                self.face2[2]
            )

        return output


class Cycle:
    def __init__(
        self,
        time: datetime,
        battery: float | None,
        temperature: float | None,
        inclination: tuple[Angle, Angle] | None
    ) -> None:
        self.time = time
        self.battery = battery
        self.temperature = temperature
        self.inclination = inclination
        self._points: dict[str, Point] = {}

    @classmethod
    def from_dict(cls, data: CycleDict) -> Cycle:
        output = cls(
            datetime.fromisoformat(data["time"]),
            data["battery"],
            data["temperature"],
            (
                Angle(data["inclination"][0]),
                Angle(data["inclination"][1])
            ) if data["inclination"] is not None else None
        )

        for p in data["points"]:
            output.add_measurement(
                p["name"],
                Face.F1,
                p["height"],
                (
                    Angle(p["face1"][0]),
                    Angle(p["face1"][1]),
                    p["face1"][2]
                )
            )
            if p.get("face2") is None:
                continue

            output.add_measurement(
                p["name"],
                Face.F2,
                p["height"],
                (
                    Angle(p["face2"][0]),
                    Angle(p["face2"][1]),
                    p["face2"][2]
                )
            )

        return output

    def add_measurement(
        self,
        ptid: str,
        face: Face,
        height: float,
        measurement: tuple[Angle, Angle, float]
    ) -> None:
        if ptid not in self._points:
            self._points[ptid] = Point(ptid, height)

        point = self._points[ptid]
        if face == Face.F1:
            if point.face1 is not None:
                raise ValueError(
                    f"Face 1 observation already exists for {ptid}"
                )

            point.face1 = measurement
            return

        if point.face2 is not None:
            raise ValueError(f"Face 2 observation already exists for {ptid}")

        point.face2 = measurement

    def to_dict(self) -> CycleDict:
        return {
            "time": self.time.isoformat(),
            "battery": self.battery,
            "inclination": (
                float(self.inclination[0]),
                float(self.inclination[1])
            ) if self.inclination is not None else None,
            "temperature": self.temperature,
            "points": [p.to_dict() for p in self._points.values()]
        }


class Session:
    def __init__(self, station: Coordinate, iheight: float) -> None:
        self.station = station
        self.iheight = iheight
        self.cycles: list[Cycle] = []

    @classmethod
    def from_dict(cls, data: SessionDict) -> Session:
        output = cls(
            Coordinate(
                data["station"][0],
                data["station"][1],
                data["station"][2]
            ),
            data["instrumentheight"]
        )

        output.cycles = [Cycle.from_dict(c) for c in data["cycles"]]

        return output

    def to_dict(self) -> SessionDict:
        return {
            "station": (
                self.station.x,
                self.station.y,
                self.station.z
            ),
            "instrumentheight": self.iheight,
            "cycles": [c.to_dict() for c in self.cycles]
        }

    def export_to_json(
        self,
        filepath: str
    ) -> None:
        make_directory(filepath)
        with open(filepath, "wt", encoding="utf8") as file:
            json.dump(
                self.to_dict(),
                file,
                indent=4
            )



import os
from datetime import datetime
from logging import getLogger
from typing import Iterator, Literal
from itertools import chain
import pathlib

from geocompy.data import Angle, Coordinate
from geocompy.communication import open_serial
from geocompy.geo import GeoCom
from geocompy.geo.gctypes import GeoComCode
from geocompy.geo.gcdata import Face

from ..utils import make_logger
from ..targets import (
    TargetPoint,
    TargetList,
    load_targets_from_json
)
from .sessions import (
    Session,
    Cycle
)


def iter_targets(
    points: TargetList,
    order: str
) -> Iterator[tuple[Face, TargetPoint]]:
    match order:
        case "AaBb":
            return ((f, t) for t in points for f in (Face.F1, Face.F2))
        case "AabB":
            return (
                (f, t) for i, t in enumerate(points)
                for f in (
                    (Face.F1, Face.F2)
                    if i % 2 == 0 else
                    (Face.F2, Face.F1)
                )
            )
        case "ABab":
            return chain(
                ((Face.F1, t) for t in points),
                ((Face.F2, t) for t in points)
            )
        case "ABba":
            return chain(
                ((Face.F1, t) for t in points),
                ((Face.F2, t) for t in reversed(points))
            )
        case "ABCD":
            return ((Face.F1, t) for t in points)

    exit(1200)


def measure_set(
    tps: GeoCom,
    filepath: str,
    order_spec: Literal['AaBb', 'AabB', 'ABab', 'ABba', 'ABCD'],
    count: int = 1,
    pointnames: str = ""
) -> Session:
    applog = getLogger("APP")
    points = load_targets_from_json(filepath)
    if pointnames != "":
        use_points = set(pointnames.split(","))
        loaded_points = set(points.get_target_names())
        excluded_points = loaded_points - use_points
        applog.debug(f"Excluding points: {excluded_points}")
        for pt in excluded_points:
            points.pop_target(pt)

    tps.aut.turn_to(0, Angle(180, 'deg'))
    incline = tps.tmc.get_angle_inclination('MEASURE').params
    temp = tps.csv.get_internal_temperature().params
    battery = tps.csv.check_power().params
    resp_station = tps.tmc.get_station().params
    if resp_station is None:
        station = Coordinate(0, 0, 0)
        iheight = 0.0
        applog.warning(
            "Could not retrieve station and instrument height, using default"
        )
    else:
        station, iheight = resp_station

    session = Session(station, iheight)
    for i in range(count):
        applog.info(f"Starting set cycle {i + 1}")
        output = Cycle(
            datetime.now(),
            battery[0] if battery is not None else None,
            temp,
            (incline[4], incline[5]) if incline is not None else None
        )

        for f, t in iter_targets(points, order_spec):
            applog.info(f"Measuring {t.name} ({f.name})")
            rel_coords = (
                (t.coords + Coordinate(0, 0, t.height))
                - (station + Coordinate(0, 0, iheight))
            )
            hz, v, _ = rel_coords.to_polar()
            if f == Face.F2:
                hz = (hz + Angle(180, 'deg')).normalized()
                v = Angle(360, 'deg') - v

            tps.aut.turn_to(hz, v)
            resp_atr = tps.aut.fine_adjust(0.5, 0.5)
            if resp_atr.error != GeoComCode.OK:
                applog.error(
                    f"ATR fine adjustment failed ({resp_atr.error.name}), "
                    "skipping point"
                )
                continue

            tps.bap.set_prism_type(t.prism)
            tps.tmc.do_measurement()
            resp_angle = tps.tmc.get_simple_measurement(10)
            if resp_angle.params is None:
                applog.error(
                    f"Error during measurement ({resp_angle.error.name}), "
                    "skipping point"
                )
                continue

            output.add_measurement(
                t.name,
                f,
                t.height,
                resp_angle.params
            )
            applog.info("Done")

        session.cycles.append(output)

    tps.aut.turn_to(0, Angle(180, 'deg'))

    return session


def main(
    port: str,
    targets: pathlib.Path,
    directory: pathlib.Path,
    baud: int = 9600,
    timeout: int = 15,
    retry: int = 1,
    sync_after_timeout: bool = False,
    format: str = "setmeasurement_{time}.json",
    cycles: int = 1,
    order: Literal['AaBb', 'AabB', 'ABab', 'ABba', 'ABCD'] = "ABba",
    sync_time: bool = True,
    points: str = "",
    debug: bool = False,
    info: bool = False,
    warning: bool = False,
    error: bool = False,
) -> None:
    log = make_logger("TPS", debug, info, warning, error)
    applog = make_logger("APP", debug, info, warning, error)
    applog.info("Starting measurement session")

    with open_serial(
        port,
        retry=retry,
        sync_after_timeout=sync_after_timeout,
        speed=baud,
        timeout=timeout
    ) as com:
        tps = GeoCom(com, log)
        if sync_time:
            tps.csv.set_datetime(datetime.now())

        session = measure_set(
            tps,
            str(targets),
            order,
            cycles,
            points
        )

    applog.info("Finished measurement session")

    timestamp = session.cycles[0].time.strftime("%Y%m%d_%H%M%S")
    filename = os.path.join(
        directory,
        format.format(time=timestamp, order=order, cycle=cycles)
    )
    session.export_to_json(filename)
    applog.info(f"Saved measurement results at '{filename}'")



from typing import Any

from click_extra import (
    extra_command,
    argument,
    option,
    file_path,
    Choice
)

from ..utils import (
    com_port_argument,
    com_option_group
)


@extra_command(
    "settings",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@argument(
    "protocol",
    help="communication protocol",
    type=Choice(["geocom", "gsidna"], case_sensitive=False)
)
@argument(
    "file",
    help="file to save settings to",
    type=file_path(readable=False)
)
@com_option_group()
@option(
    "-f",
    "--format",
    help="settings file format",
    type=Choice(["auto", "json", "yaml", "toml"], case_sensitive=False),
    default="auto"
)
@option(
    "--defaults",
    help=(
        "add defaults for settings that could not be saved "
        "(e.g. not applicable to the current instrument)"
    ),
    is_flag=True
)
def cli_download(**kwargs: Any) -> None:
    """Save instrument settings to file."""
    from .save import main

    main(**kwargs)


@extra_command(
    "settings",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@argument(
    "settings",
    help="file containing instrument settings",
    type=file_path(exists=True, readable=True)
)
@com_option_group()
@option(
    "-f",
    "--format",
    help="settings file format",
    type=Choice(["auto", "json", "yaml", "toml"], case_sensitive=False),
    default="auto"
)
def cli_upload(**kwargs: Any) -> None:
    """Load instrument settings from file."""
    from .load import main

    main(**kwargs)


@extra_command(
    "settings",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@argument(
    "file",
    help="settings file to validate",
    type=file_path(exists=True, readable=True)
)
@option(
    "-f",
    "--format",
    help="settings file format",
    type=Choice(["auto", "json", "yaml", "toml"], case_sensitive=False),
    default="auto"
)
def cli_validate(**kwargs: Any) -> None:
    """Validate instrument settings config."""
    from .validate import main

    main(**kwargs)



from click_extra import extra_command


@extra_command("terminal", params=None)  # type: ignore[misc]
def cli() -> None:
    """Test instrument commands in an interactive TUI."""

    from .app import GeoComTerminal

    tui = GeoComTerminal()
    tui.run()



from typing import Any

from click_extra import extra_command

from ..utils import (
    com_option_group,
    com_port_argument
)


@extra_command(
    "geocom",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@com_option_group()
def cli_geocom(**kwargs: Any) -> None:
    """Test the availability of various GeoCom protocol functions on an
    instrument."""
    from .app import main

    kwargs["protocol"] = "geocom"
    main(**kwargs)


@extra_command(
    "gsidna",
    params=None,
    context_settings={"auto_envvar_prefix": None}
)  # type: ignore[misc]
@com_port_argument()
@com_option_group()
def cli_gsidna(**kwargs: Any) -> None:
    """Test the availability of various GSI Online DNA functions on an
    instrument."""
    from .app import main

    kwargs["protocol"] = "gsidna"
    main(**kwargs)



from __future__ import annotations

import os
import json
import re
from typing import TypedDict, Iterator, Any

from jsonschema import validate
from geocompy.data import Coordinate
from geocompy.geo.gcdata import Prism

from .utils import make_directory


class TargetPointDict(TypedDict):
    name: str
    prism: str
    height: float
    coords: tuple[float, float, float]


class TargetListDict(TypedDict):
    targets: list[TargetPointDict]


class TargetPoint:
    def __init__(
        self,
        name: str,
        prism: Prism,
        height: float,
        coords: Coordinate
    ) -> None:
        self.name = name
        self.prism = prism
        self.height = height
        self.coords = coords

    @classmethod
    def from_dict(cls, data: TargetPointDict) -> TargetPoint:
        return cls(
            data["name"],
            Prism[data["prism"]],
            data["height"],
            Coordinate(*data["coords"])
        )

    def to_dict(self) -> TargetPointDict:
        return {
            "name": self.name,
            "prism": self.prism.name,
            "height": self.height,
            "coords": (self.coords.x, self.coords.y, self.coords.z)
        }

    def __str__(self) -> str:
        return str(self.to_dict())


class TargetList:
    def __init__(self) -> None:
        self._targets: list[TargetPoint] = []
        self._targets_lookup: dict[str, TargetPoint] = {}

    @classmethod
    def from_dict(cls, data: TargetListDict) -> TargetList:
        output = cls()
        for item in data["targets"]:
            point = TargetPoint.from_dict(item)
            output._targets.append(point)
            output._targets_lookup[point.name] = point

        return output

    def to_dict(self) -> TargetListDict:
        return {"targets": [t.to_dict() for t in self._targets]}

    def __str__(self) -> str:
        return str(self.to_dict())

    def __contains__(self, name: str) -> bool:
        return name in self._targets_lookup

    def __len__(self) -> int:
        return len(self._targets)

    def __iter__(self) -> Iterator[TargetPoint]:
        return iter(self._targets)

    def __reversed__(self) -> Iterator[TargetPoint]:
        return reversed(self._targets)

    def add_target(self, target: TargetPoint) -> None:
        if target.name in self._targets_lookup:
            raise ValueError(f"Target {target.name} already exists")

        self._targets.append(target)
        self._targets_lookup[target.name] = target

    def pop_target(self, name: str) -> TargetPoint:
        target = self._targets_lookup[name]
        self._targets.remove(target)
        self._targets_lookup.pop(name)
        return target

    def get_target(self, name: str) -> TargetPoint:
        return self._targets_lookup[name]

    def get_target_names(self) -> list[str]:
        return list(self._targets_lookup.keys())


def export_targets_to_json(filepath: str, targets: TargetList) -> None:
    make_directory(filepath)
    with open(filepath, "wt", encoding="utf8") as file:
        json.dump(targets.to_dict(), file, indent=4)


def load_targets_from_json(filepath: str) -> TargetList:
    with (
        open(filepath, "rt", encoding="utf8") as file,
        open(
            os.path.join(
                os.path.dirname(__file__),
                "schema_targets.json"
            ),
            "rt",
            encoding="utf8"
        ) as file_schema
    ):
        data: TargetListDict = json.load(file)
        schema: dict[str, Any] = json.load(file_schema)

    validate(data, schema)

    return TargetList.from_dict(data)


_COLUMNS = re.compile(r"^[PENZ_]{4,}$")


def import_targets_from_csv(
    filepath: str,
    separator: str,
    columns: str,
    reflector: Prism,
    header: int = 0
) -> TargetList:
    columns = columns.upper()

    if not _COLUMNS.match(columns):
        raise ValueError("Invalid column spec")

    col_ptid = columns.find("P")
    col_east = columns.find("E")
    col_north = columns.find("N")
    col_height = columns.find("Z")

    if -1 in (col_ptid, col_east, col_north, col_height):
        raise ValueError(
            "Some of the PENZ columns were not found in the column spec"
        )

    output = TargetList()

    with open(filepath, "rt", encoding="utf8") as file:
        for i in range(header):
            next(file)

        for line in file:
            fields = line.strip().split(separator)
            point = TargetPoint(
                fields[col_ptid],
                reflector,
                0.0,
                Coordinate(
                    float(fields[col_east]),
                    float(fields[col_north]),
                    float(fields[col_height])
                )
            )
            output.add_target(point)

    return output



from typing import Any, cast, TypedDict
from pathlib import Path

import json
import yaml
import toml

from ..utils import echo_red


class SubsystemSettingsDict(TypedDict):
    subsystem: str
    options: dict[str, Any]


class SettingsDict(TypedDict):
    protocol: str
    settings: list[SubsystemSettingsDict]


def read_settings(
    file: Path,
    format: str = "auto"
) -> SettingsDict:
    if format == "auto":
        format = file.suffix[1:].lower() if file.suffix else ""

    match format:
        case "json":
            with file.open("rt", encoding="utf8") as settings:
                data = cast(SettingsDict, json.load(settings))
        case "yaml" | "yml":
            with file.open("rt", encoding="utf8") as settings:
                data = cast(SettingsDict, yaml.load(settings, yaml.Loader))
        case "toml":
            # The TOML package doesn't support heterogenous arrays, even tho it
            # was added to the language spec in 2019. Therefore the standard
            # lib tomllib/tomli has to be used for parsing.
            import tomllib as toml
            with file.open("rb") as settings:
                data = cast(SettingsDict, toml.load(settings))
        case _:
            echo_red(f"Unknown file format: {format}")
            exit(1)

    return data


def write_settings(
    data: SettingsDict,
    file: Path,
    format: str = "auto"
) -> None:
    if format == "auto":
        format = file.suffix[1:].lower() if file.suffix else ""

    with file.open("wt", encoding="utf8") as settings:
        match format:
            case "json":
                json.dump(data, settings, indent=4)
            case "yaml" | "yml":
                yaml.dump(data, settings, yaml.Dumper)
            case "toml":
                toml.dump(data, settings)
            case _:
                echo_red(f"Unknown file format: {format}")
                exit(1)



from __future__ import annotations

from pathlib import Path
from typing import Any, Callable

from geocompy.communication import open_serial
from geocompy.geo import GeoCom
from geocompy.geo.gctypes import GeoComSubsystem, GeoComResponse, GeoComCode
from geocompy.gsi.dna import GsiOnlineDNA
from geocompy.gsi.gsitypes import GsiOnlineResponse
from geocompy.gsi.dna.settings import GsiOnlineDNASettings

from ..utils import echo_red, echo_yellow, echo_green
from .io import read_settings, SettingsDict
from .validate import validate_settings


def set_setting_geocom(
    system: GeoComSubsystem,
    setting: str,
    value: int | float | bool | str | list[int | float | bool | str]
) -> None:
    if isinstance(value, bool):
        name = f"switch_{setting}"
    else:
        name = f"set_{setting}"

    method: Callable[
        ...,
        GeoComResponse[Any]
    ] | None = getattr(system, name, None)
    if method is None:
        echo_yellow(f"Could not find '{name}' to set '{setting}'")
        return

    if isinstance(value, list):
        response = method(*value)
    else:
        response = method(value)

    if response.error != GeoComCode.OK:
        echo_yellow(f"Could not set '{setting}' ({response.error.name})")
        return


def set_setting_gsidna(
    system: GsiOnlineDNASettings,
    setting: str,
    value: int | float | bool | str
) -> None:
    name = f"set_{setting}"
    method: Callable[
        ...,
        GsiOnlineResponse[bool]
    ] | None = getattr(system, name, None)
    if method is None:
        echo_yellow(f"Could not find '{name}' to set '{setting}'")
        return

    response = method(value)

    if response.value is None or not response.value:
        echo_yellow(f"Could not set '{setting}' ({response.response})")
        return


def upload_settings_geocom(
    protocol: GeoCom,
    settings: SettingsDict
) -> None:
    for item in settings["settings"]:
        sysname = item["subsystem"]
        subsystem: Any = getattr(protocol, sysname)
        if subsystem is None:
            echo_red(f"Could not find '{sysname}' subsystem")
            exit(1)

        for option, value in item["options"].items():
            if value is None:
                continue

            set_setting_geocom(subsystem, option, value)


def upload_settings_gsidna(
    protocol: GsiOnlineDNA,
    settings: SettingsDict
) -> None:
    for item in settings["settings"]:
        sysname = item["subsystem"]
        subsystem: Any = getattr(protocol, sysname)
        if subsystem is None:
            echo_red(f"Could not find '{sysname}' subsystem")
            exit(1)

        for option, value in item["options"].items():
            if value is None:
                continue

            set_setting_gsidna(subsystem, option, value)


def main(
    port: str,
    settings: Path,
    baud: int = 9600,
    timeout: int = 15,
    retry: int = 1,
    sync_after_timeout: bool = False,
    format: str = "auto"
) -> None:
    data = read_settings(settings, format)
    if not validate_settings(data):
        echo_red("Settings file does not follow schema")
        exit(1)

    with open_serial(
        port,
        retry=retry,
        sync_after_timeout=sync_after_timeout,
        speed=baud,
        timeout=timeout
    ) as com:
        match data["protocol"]:
            case "geocom":
                tps = GeoCom(com)
                upload_settings_geocom(tps, data)
            case "gsidna":
                dna = GsiOnlineDNA(com)
                upload_settings_gsidna(dna, data)
            case _:
                echo_red(f"Unknown protocol: {data["protocol"]}")
                exit(1)

        echo_green(f"Settings loaded from {settings}")
